@@ -35,8 +35,49 @@ class Snaplib:
     '''
     data preprocessing library
     '''
+    
+    def __init__(self, encoder_pool=dict(), decoder_pool=dict(), encoded_columns=list()):
+        self.encoder_pool = encoder_pool
+        self.decoder_pool = decoder_pool
+        self.encoded_columns = encoded_columns
+    
 
+    
+    
+    @property
+    def encoder_pool(self):
+        return self.__encoder_pool
 
+    @encoder_pool.setter
+    def encoder_pool(self, encoder_pool):
+        self.__encoder_pool = encoder_pool
+    
+    
+    
+    @property
+    def decoder_pool(self):
+        return self.__decoder_pool
+
+    @decoder_pool.setter
+    def decoder_pool(self, decoder_pool):
+        self.__decoder_pool = decoder_pool
+    
+    
+    
+    @property
+    def encoded_columns(self):
+        return self.__encoded_columns
+
+    @encoded_columns.setter
+    def encoded_columns(self, encoded_columns):
+        self.__encoded_columns = encoded_columns
+        
+        
+    
+    
+    
+        
+        
 
 
     def nan_info(self, df):
@@ -110,21 +151,27 @@ class Snaplib:
         df.reset_index(drop=True, inplace=True)
         dr_dupl_shape = df.shape
         if verbose:
-            print(f'{start_shape[0] - dr_dupl_shape[0]} - rows have been dropped')
+            print(f'{start_shape[0] - dr_dupl_shape[0]} rows have been dropped')
             print(f'shape: {dr_dupl_shape}\n')
 
         # DROP COLUMNS with 1 unique value
             print('DROP COLUMNS with 1 unique value:')
+            
+            count_drop_columns = 0
         for col in df.columns:
-            unique_array = df[col].unique()
+            unique_array = np.array(df[col].unique())
+
             if len(unique_array) == 1:
+                count_drop_columns += 1
                 df.drop([col], inplace=True, axis=1)
                 if verbose:
-                    print(f'column "{col}" cnontains 1 unique value -> have been dropped')
-            elif len(unique_array) == 2 and np.any(np.isnan(unique_array)):
+                    print(f'column "{col}" cnontains 1 unique value - has been dropped')
+            elif len(unique_array) == 2 and np.any(pd.isnull(df[col])):
                 if verbose:
                     print(f'!!! column "{col}" cnontains 1 unique value and np.nan')
+        
         if verbose:
+            print(f'{count_drop_columns} columns have been dropped')
             print(f'shape: {df.shape}\n')
 
         # DROP ROWS with NaN IN TARGET
@@ -134,7 +181,7 @@ class Snaplib:
             nan = df[df[target].isnull()]
             indeces = list(nan.index)
             if verbose:
-                print(f'{len(indeces)} - rows have been dropped')
+                print(f'{len(indeces)} rows have been dropped')
             df = df.drop(df.index[indeces])
             df.reset_index(drop=True, inplace=True)
         if verbose:
@@ -143,11 +190,313 @@ class Snaplib:
             self.nan_plot(df)
         return df
 
+    
+    
+    
+    
+    
+    
+    
+    
+    
+    
+    
+    def encode_column(self, df, column):
+        '''
+        encode one column in dataframe
+
+        Use case:
+        df = Snaplib.encode_column(df, column_name_str)
+        '''
+        self.encoder_pool[column] = {}
+        self.decoder_pool[column] = {}
+        not_nan_index=df[df[column].notnull()].index
+        values_set = list(set(list(df.loc[not_nan_index, column])))
+        value = 0.0
+        for el in values_set:
+            self.encoder_pool[column][el] = value
+            self.decoder_pool[column][value] = el
+            value += 1
+        df[column] = df[column].map(self.encoder_pool[column])
+        df[column] = df[column].astype('float64')
+        return df
+    
+    
+    
+    
+    
+    
+    
+    
+    
+    def decode_column(self, df, column):
+        '''
+        decode one column in dataframe
+
+        Use case:
+        df = Snaplib.decode_column(df, column_name_str)
+        '''
+        df[column] = df[column].map(self.decoder_pool[column])
+        df[column] = df[column].astype('object')
+        return df
+    
+    
+    
+    
+    
+    
+    
 
 
 
+    def encode_dataframe(self, df):
+        '''
+        encode a dataframe
+
+        Use case:
+        df = Snaplib.encode_dataframe(df)
+        '''
+        self.encoder_pool = {}
+        self.decoder_pool = {}
+        self.encoded_columns = []
+
+        types = df.dtypes
+        for col in df.columns:
+            feature_type = types[col]
+            if feature_type == 'object' or feature_type == 'bool':            
+                df[col] = self.encode_column(df[[col]], col)
+                self.encoded_columns.append(col)
+            else:
+                # object type column with NaN sometimes has type float64
+                try:
+                    df.loc[:, col] = df.loc[:, col] + 0
+                except:
+                    df[col] = df[col].astype('object')
+                    df[col] = self.encode_column(df[[col]], col)
+                    self.encoded_columns.append(col)
+        return df
+    
+    
+    
+    
+    
+    
+    
+    
+    
+
+    
+    def decode_dataframe(self, df):
+        '''
+        decode a dataframe
+
+        Use case:
+        df = Snaplib.decode_dataframe(df)
+        '''
+        for col in self.encoded_columns:
+            df[col] = self.decode_column(df[[col]], col)
+        return df
 
 
+    
+    
+    
+    
+    
+    
+    
+    
+    
+    def k_folds_split(self, df, target_name, k):
+
+        k_fold_dict = { 
+                        'train_X' : [],
+                        'test_X'  : [], 
+                        'train_y' : [],
+                        'test_y'  : [],
+                      }
+
+
+        for i in range(0, k):
+            train_X, test_X, train_y, test_y = self.train_test_split_balanced(df, 
+                                                                              target_name, 
+                                                                              random_state=i, 
+                                                                              test_size=1/k, 
+                                                                              research_iter=0
+                                                                              )
+            k_fold_dict['train_X'].append(train_X)
+            k_fold_dict['test_X'].append(test_X)
+            k_fold_dict['train_y'].append(train_y)
+            k_fold_dict['test_y'].append(test_y)
+
+        return k_fold_dict
+
+
+    
+    
+    
+    
+    
+    
+    
+    
+    
+    
+    
+    def predict_stacked(self, 
+                        all_algorithms, 
+                        X__train, 
+                        y__train, 
+                        X__pred, 
+                        y__test=None, 
+                        task='clsf', 
+                        verbose=0
+                        ):
+        
+        ''' 
+        Prediction method for list of algorithms.
+        
+        Use case:
+        y_hat = Snaplib.predict_stacked(
+                                        all_algorithms, 
+                                        X_train, 
+                                        y_train, 
+                                        X_pred, 
+                                        y_test or None, 
+                                        task='clsf' or 'regr', 
+                                        verbose= 0, 1, 2
+                                        ):
+        
+        all_algorithms = list of algorithms [LGBMClassifier(), XGBClassifier(), CatBoostClassifier()].
+        X_train and y_train are data for training list of algorithms.
+        X_pred is dataframe for prediction.
+        y_test optionaly. If exist visualize it as last column on a plot (verbose=2). 
+        task='clsf' or 'regr', classification or regression
+        verbose = 0 mute, 1 verbose.
+
+        '''
+        stacked_predicts = pd.DataFrame()
+        alg_names = []
+        for alg in all_algorithms:
+            alg_name = alg.__class__.__name__[:3]
+            model = alg.fit(X__train, y__train)
+            y_hat = model.predict(X__pred)
+            if task =='clsf':
+                stacked_predicts[alg_name] = y_hat.astype('int64')
+            else:
+                stacked_predicts[alg_name] = y_hat
+            alg_names.append(alg_name)
+
+        if task =='clsf':
+            stacked_predicts['Y_HAT_STACKED'] = stacked_predicts[alg_names].mode(axis=1)[0].astype('int64')
+            if y__test is not None:
+                stacked_predicts['Y_TEST'] = y__test.values.astype('int64')
+        else:
+            stacked_predicts['Y_HAT_STACKED'] = stacked_predicts[alg_names].mean(axis=1)
+            if y__test is not None:
+                stacked_predicts['Y_TEST'] = y__test.values
+
+
+        y_hat = stacked_predicts.loc[:, 'Y_HAT_STACKED']
+        if verbose:
+            plt.figure(figsize=(5, 10))
+            sns.heatmap(stacked_predicts[-1000:], cbar=False)
+            plt.show()
+        return y_hat
+
+
+    
+    
+    
+    
+    
+    
+    
+
+
+
+    def cross_val(self, algorithms, k_fold_dict, metric, task, cv, verbose=0):
+        
+        ''' 
+        Cross Validation method for list of algorithms.
+        
+        Use case:
+        y_hat = Snaplib.cross_val(algorithms, k_fold_dict, metric, task, cv, verbose=0):
+        
+        all_algorithms = list of algorithms like [LGBMClassifier(), XGBClassifier(), CatBoostClassifier()].
+        k_fold_dict is a dictionary with the structure:
+        
+        K_FOLD = 3
+        k_fold_dict = { 
+                        'train_X' : [df_0, df_1, df_2],
+                        'test_X'  : [df_0, df_1, df_2],
+                        'train_y' : [seri_0, seri_1, seri_2],
+                        'test_y'  : [seri_0, seri_1, seri_2],
+                       }
+              
+        metric is a metric like f1_score or mean_absolute_error.
+        task='clsf' or 'regr', classification or regression.
+        cv is num K_FOLD integer 
+        verbose = 0 mute, 1 verbose.
+
+        '''
+
+        results=[]
+        test_y_all = np.array([])
+        pred_all = np.array([])
+        if task == 'clsf':
+            cm_base = np.array([[0, 0], [0, 0]])
+
+
+        for k in range(0, cv):
+            if len(algorithms) > 1:
+                pred = self.predict_stacked(algorithms, 
+                                            k_fold_dict['train_X'][k], 
+                                            k_fold_dict['train_y'][k], 
+                                            k_fold_dict['test_X'][k], 
+                                            k_fold_dict['test_y'][k], 
+                                            task,
+                                            verbose
+                                           )
+            else:
+                alg = algorithms[0]
+                alg.fit(k_fold_dict['train_X'][k], k_fold_dict['train_y'][k])
+                pred = alg.predict(k_fold_dict['test_X'][k])
+
+            if verbose:
+                if task == 'clsf':
+                    cm = confusion_matrix(k_fold_dict['test_y'][k], pred)
+                    cm_base = cm_base + cm
+
+            test_y_all = np.concatenate(([test_y_all, k_fold_dict['test_y'][k]]), axis=None)
+            pred_all = np.concatenate((pred_all, pred), axis=None)
+
+            score = metric(k_fold_dict['test_y'][k], pred)
+            results.append(score)
+
+        results_std = np.std(results)
+        score = sum(results) / len(results)
+
+        if verbose:
+            print(*algorithms, sep='\n')
+            print('')
+            print("%s %0.6f (std: +/- %0.2f)" % (metric_name, score, results_std))
+            print('\n', results, '\n')
+
+            if task == 'clsf':
+                print('\n', classification_report(test_y_all, pred_all), '\n')
+                plt.figure(figsize=(3, 3))
+                sns.heatmap(cm_base, annot=True, cmap="Blues", fmt='.0f',  cbar=False)
+                plt.show()
+
+        return score
+    
+    
+    
+    
+    
+    
+    
 
 
 
@@ -372,9 +721,7 @@ class Snaplib:
         And ensemble decrise train/test leakage.
         '''
             
-        encoder_pool = {}
-        decoder_pool = {}
-        encoded_columns = []
+
         counter_predicted_values = 0
         CLASS_VALUE_COUNTS = 20
         miss_indeces = None
@@ -382,22 +729,8 @@ class Snaplib:
 
         
         
-        def encode_column(df, enc_pool, dec_pool, column):
-            enc_pool[column] = {}
-            dec_pool[column] = {}
-            not_nan_index=df[df[column].notnull()].index
-            values_set = list(set(list(df.loc[not_nan_index, column])))
-            value = 0.0
-            for el in values_set:
-                enc_pool[column][el] = value
-                dec_pool[column][value] = el
-                value += 1
-            df[column] = df[column].map(enc_pool[column])
-            df[column] = df[column].astype('float64')
-            return df
         
-        
-        def get_columns(columns, target_column):
+        def get_predictors(columns, target_column):
             columns_now = columns[:]
             if target_column in columns:
                 columns_now.remove(target_column)
@@ -456,8 +789,6 @@ class Snaplib:
                 y_hat = stacked_predicts.loc[:, 'y_hat_final']
                 y_hat[y_hat == -np.inf] = 0
                 y_hat[y_hat == np.inf] = 0
-    #         print(stacked_predicts)
-            # del stacked_predicts
             return y_hat
 
         
@@ -466,13 +797,7 @@ class Snaplib:
             return df
         
         
-        def decode_column(df, dec_pool, column):
-            df[column] = df[column].map(dec_pool[column])
-            df[column] = df[column].astype('object')
-            return df
-        
-        
-        
+
         
         # main
         init_time = datetime.datetime.now()
@@ -492,22 +817,8 @@ class Snaplib:
         
         all_miss_features = list(data_info.index[data_info['NaN_counts'] > 0])
         
-                
         # a simple encoding
-        for col in df.columns:
-            feature_type = data_info.loc[col, 'col_type']
-            if feature_type == 'object' or feature_type == 'bool':            
-                df[col] = encode_column(df[[col]], encoder_pool, decoder_pool, col)
-                encoded_columns.append(col)
-            else:
-                # object column with NaN sometimes has type float64
-                try:
-                    df.loc[:, col] = df.loc[:, col] + 0
-                except:
-                    df[col] = df[col].astype('object')
-                    df[col] = encode_column(df[[col]], encoder_pool, decoder_pool, col)
-                    encoded_columns.append(col)
-        
+        df = self.encode_dataframe(df)
         
         # get continuous & discrete features
         continuous_features = []
@@ -529,8 +840,8 @@ class Snaplib:
             predictors = all_features[:]
             predictors.remove(target_now)
             
-            continuous_features_now = get_columns(continuous_features, target_now)
-            # discrete_features_now = get_columns(discrete_features, target_now)
+            continuous_features_now = get_predictors(continuous_features, target_now)
+            # discrete_features_now = get_predictors(discrete_features, target_now)
             
             # indexes of missing data in target_now (data for prediction)
             miss_indeces = list((df[pd.isnull(df[target_now])]).index)
@@ -626,7 +937,6 @@ class Snaplib:
                     print(f'Classification Report:\n')
                     print(classification_report(y_test, pred_test), '\n')
 
-    #             ADVANCED  alg = LGBMClassifier(n_jobs=-1, random_state=0, class_weight=class_weight)                
                 pred_miss = predict(X_train, y_train, miss_df, algorithms, len_target_values_counted)
 
                 pred_miss = [int(i) for i in pred_miss]
@@ -662,7 +972,6 @@ class Snaplib:
                     print(f'min for {target_now}: {df[target_now].min()}')
                     print(f'avg for {target_now}: {df[target_now].mean()}')
                     print(f'max for {target_now}: {df[target_now].max()}\n')
-                    
 
 
     #             ADVANCED  alg = LGBMRegressor(n_jobs=-1, random_state=0)        
@@ -674,7 +983,6 @@ class Snaplib:
                 imput_missing_value_to_main_df(df, miss_indeces, list(pred_miss), target_now)
                 counter_predicted_values += len(miss_indeces)
                 
-                
 
             else:
                 if verbose:
@@ -700,9 +1008,8 @@ class Snaplib:
                 print(f'Required time:  {str(requared)}\n')
 
             
-        # return states to their initial states
-        for col in encoded_columns:
-            df[col] = decode_column(df[[col]], decoder_pool, col)
+        # return states to their initial states        
+        df = self.decode_dataframe(df)
             
         for col in df.columns:
             df[col] = df[col].astype(data_info.loc[col, 'col_type'])
